;-----------------------------------------------------------------------------
; cc.asm
; Copyright 1994, 1995, 1996, 1997, 2003, 2004 Eric Smith & Richard Ottosen
;
; Closed Caption Decoder using PIC16F628A running at 20 MHz
; requires EL4581C sync separator or equivalent
;
; Assembles with gpasm 0.11.8 alpha, part of gputils package.  Requires
; the "-i" option for case insensitivity.
;     http://gputils.sourceforge.net/
;
; Also assembles with Microchip MPLAB/MPASM 6.40.  Requires the "/c" option
; for case insensitivity.
;
; $Id: cc.asm,v 1.23 2004/01/04 22:12:46 eric Exp eric $
;-----------------------------------------------------------------------------

	processor pic16f648a
	include	p16f648a.inc

	radix	dec

	__config	_BOREN_ON&_CP_OFF&_DATA_CP_OFF&_PWRTE_ON&_WDT_OFF&_LVP_OFF&_MCLRE_ON&_HS_OSC


majority    equ	1	; use majority voting
has_dbg	    equ	1
has_pwm_inv equ	1
has_pzt	    equ	0

startup_dbg equ 0	; output a '*' on startup

dbg_echo    equ 0	; echo commands

dcrest_dbg  equ	0	; use DCD output as DC restore timing debug
			; if true, DCD will go high 200 ns after DC restore
			;   starts, and go low 200 ns before it ends.

samp_dbg    equ	0	; use DCD output as sample timing debug
			; if true, DCD will go high 200 ns before first sample,
			;   and go low 400 ns after last sample


fosc	 equ	20000000	; 20 MHz
ser_rate equ	19200
;ser_rate equ	115200		; fast for debug


brg_val		equ	(((2*fosc)/(16*ser_rate))+1)/2-1

rx_ring_size	equ	16
tx_ring_size	equ	48
cmd_buf_size	equ	16

	errorlevel -305		; don't complain about missing ",f"
	errorlevel -224		; don't complain about tris and
				;   option instructions

;-----------------------------------------------------------------------------
; specific register bit assignments for the closed-caption decoder
;-----------------------------------------------------------------------------

; I/O port A:

peak	equ	0		;   bit 0 = peak detector reset open-drain output
				;   bit 1 = comparator input
restore	equ	2		;   bit 2 = back porch DC restore open-drain output
c1outpin equ	3		;   bit 3 = comparator output
c2outpin equ	4		;   bit 4 = comparator output, open drain

; values for "tris porta"
tris_a_no_clamp	  equ	0ffh-((1<<c1outpin)|(1<<c2outpin))
				; neither DC restore clamp nor peak detect clamp				; c1out and c2out must always be outputs

tris_a_clamp	  equ	tris_a_no_clamp-((1<<peak)|(1<<restore))
				; both DC Restore clamp and peak detect clamp

tris_a_no_dc_rest equ	tris_a_no_clamp-(1<<peak)	; Peak detect reset on


; I/O port B:

csync	equ	0		;   bit 0 = composite sync input from sync
				;           separator (active low) (INT)

rxd	equ	1		;   bit 1 = rs232 serial data input (UART)

txd	equ	2		;   bit 2 = rs232 serial data output (UART)

	if	has_pwm_inv
pump	equ	3		;   bit 3 = Charge pump drive for RS232
				;           -4 Volts (PWM, timer 2)
	endif

pzt	equ	4		;   bit 4 = pzt speaker output
				; (must be low when using ICD2 with
				;  PIC16F648A-ICD bond-out chip)

dcd	equ	5		;   bit 5 = rs232 Data Carrier Detect output
				;           and active caption LED

vsync	equ	6		;   bit 6 = vertical sync input from sync
				;           separator (active low)

oddf	equ	7		;   bit 7 = odd field input from sync
				;           separator (active high)


tris_b_val	equ	(1<<vsync)|(1<<rxd)|(1<<txd)|(1<<csync)|(1<<oddf)

; One wouldn't expect to set the USART TxD pin as an input, but the
; USART chapter in the PIC16F628A data sheet says it is necessary
; (DS40044A, section 12.0 on p. 69).  On the other hand, Figure 5-10
; "Block diagram of RB1/TX/CK pin" (p. 38) appears to show that the
; peripheral enable will override the TRIS bit.  In practice it doesn't
; seem to matter whether we set the TRIS bit or not, so perhaps the
; diagram is correct.  Rich suggests that the text in section 12.0 may
; be leftover from earlier parts that behaved differently.


datap	equ	cmcon
datab	equ	7


;-----------------------------------------------------------------------------
; misc macros
;-----------------------------------------------------------------------------

fallto	macro	target
	list	x=off
	if	$ != target
	error	"fallto label must follow immediately"
	endif
	list	x=on
	endm


;-----------------------------------------------------------------------------
; ring buffer macro
;-----------------------------------------------------------------------------

; advance buffer pointer, wrapping if necessary

; This is a simplified version of the macro from uarttest.  Unfortunately,
; gpasm doesn't like the optimizations of the uarttest version.

ringadv	macro	ptr,base,size
	incf	ptr
	movf	ptr,w
	xorlw	base+size
	movlw	base
	btfsc	status,z
	movwf	ptr
	endm


;-----------------------------------------------------------------------------
; ascii characters
;-----------------------------------------------------------------------------

bel	equ	07h		; bell
bs	equ	08h		; back space
tab	equ	09h		; horizontal tab
lf	equ	0ah		; line feed
ff	equ	0ch		; form feed
cr	equ	0dh		; carriage return
xon	equ	11h		; transmit on
xoff	equ	13h		; transmit off
eof	equ	1ah		; end of file
esc	equ	1bh		; escape
sp	equ	20h		; space


;-----------------------------------------------------------------------------
; ram
;-----------------------------------------------------------------------------

rambase	equ	20h
ramb1	equ	rambase+80h
intram	equ	71h		; 70h-7fh mirrored in all banks.  We avoid
				;   70h because it is used by the ICD2.

	org	rambase

count:	res	1		; general purpose counter
dlycnt:	res	1		; counter for delays (used in several places)
mscnt:	res	1		; number of milliseconds using delay routine
cycles:	res	1		; number of cycles in a beep
period:	res	1		; period of a click
cmdptr:	res	1		; pointer into command table

serreg:	res	1		; serial output character buffer
bitcnt:	res	1		; serial output bit count
speed:	res	1		; serial speed

temp:	res	1		; very temporary storage
temp2:	res	1
temp3:	res	1

cmd_buf_ip:	res	1	; command buffer input pointer
cmd_buf_op:	res	1	; command buffer output pointer

rx_ring_ip:	res	1
rx_ring_op:	res	1
rx_ring_cnt:	res	1

tx_ring_ip:	res	1
tx_ring_op:	res	1
tx_ring_cnt:	res	1

prev_portb:	res	1	; saves value of port b inputs on previous
				;   scan line, used to detect falling edge
				;   of vsync

intflags: res	1		; interrupt flags
even_field     equ 0		;   current field is even field
odd_field      equ 1		;   current field is odd field
hold_time_code equ 2		;   hold frame count (being set in main)


mflags:	res	1		; misc. flags
lazy_cr	  equ	0		;   lazy cr pending
last_bit  equ	1		;   last bit (used in getsbit)


sflags:	res	1		; sample flags
s_even_field equ 0		;   sample taken on even field
s_odd_field  equ 1		;   sample taken on odd field
s_ready	     equ 2		;   sample ready
s_overrun    equ 3		;   overrun


capflags: res 1			; capture enable flags
cap_even_field equ 0		;   capture even field
cap_odd_field  equ 1		;   capture odd field


mode:	res	1		; mode flags
fcmode	equ	1		;   bit 1 = frame count enable
nondrop	equ	2		;   bit 2 = non-drop time code               
rawmode	equ	3		;   bit 3 = raw mode
showunk	equ	4		;   bit 4 = show unknown control codes in hex
allerr	equ	5		;   bit 5 = show all errors
run	equ	6		;   bit 6 = run
dbgmode	equ	7		;   bit 7 = debug mode


line_cnt res	1		; video line counter

acc:	res	5		; main accumulator (one extra byte needed for
				;   div_10)
acc2:	res	4		; another accumulator

tc_frame: res	1		; frame, seconds, minutes, hours
tc_sec:	res	1
tc_min:	res	1
tc_hr:	res	1

samp_tc_frame: res 1		; time_code at which sample data was
samp_tc_sec:   res 1		;   captured
samp_tc_min:   res 1
samp_tc_hr:    res 1

sidx8:	res	1
sidx1:	res	1

trcnt:	res	1		; count of run-in transitions detected
errflg:	res	1		; error flag, zero if no error

odata:	res	2		; decoded data bytes

samp_count equ	128
samp_bytes equ	(samp_count+7)/8

sample:	res	samp_bytes	; 136 samples at 2.5 MHz (400 ns interval).
				; caption data is just under 130 samples long.
				; Extra samples allow some slop for delay to
				; start of run-in.

	org	ramb1
rx_ring:	res	rx_ring_size
tx_ring:	res	tx_ring_size
cmd_buf:	res	cmd_buf_size

	org	intram
irq_w:		res	1
irq_stat:	res	1
irq_fsr:		res	1

tx_temp:	res	1
tx_fsr_save:	res	1


;-----------------------------------------------------------------------------
; reset and interrupt vectors
;-----------------------------------------------------------------------------

	org	0

	goto	reset


;-----------------------------------------------------------------------------
; short delay subroutines
; there is only room here (between reset and interrupt vectors) for three
; instructions
;-----------------------------------------------------------------------------

delay6:	nop
delay5:	nop
delay4:	return


;-----------------------------------------------------------------------------
; ineterrupt handler
; sources:
;   external interrupt (RB0) - composite sync falling edge
;   timer 1 - csync timeout if no csync falling edge in 120 us
; Serial (UART) receive and transmit do not generate interrupts, to avoid
; adding latency to csync interrupt handling, but are polled on all
; interrupts except line 21.
;-----------------------------------------------------------------------------

	if	($ > 0004h)
	error	"interrupt handler must be at address 4"
	endif

	org	0004h

	movwf	irq_w		;  0  save context:	 w, status, fsr
	swapf	status,w	;  1
	movwf	irq_stat	;  2
	movf	fsr,w		;  3
	movwf	irq_fsr		;  4
	clrf	status		;  5  clear rp1..0, irp
	
	btfss	intcon,intf	;  6 csync interrupt?
	goto	check_uart	;  7 no, must be timeout

	bcf	intcon,intf	;  8 clear interrupt

	incf	line_cnt	;  9 increment line counter

	movf	portb,w		; 10 save portb for next time (as close to the
				;      following btfss instruction as possible)
	btfss	portb,vsync	; 11 if vsync is low
	btfss	prev_portb,vsync ; 12 and was high last time
	goto	not_vsync_start	; 13
	goto	vsync_start	; 14

not_vsync_start:
	movwf	prev_portb	; 15

	movf	line_cnt,w	; 16 are we on line 21?
	xorlw	21		; 17
	btfsc	status,z	; 18
	goto	line21		; 19 yes, go capture samples (timing too tight
				;      to also DC restore on line 21)

	clrf	porta		; 20 Make sure output is still low
	movlw	tris_a_clamp	; 21 prepare to clamp video for DC restore

	btfss	portb,csync	; 22 has csync returned high?
	goto	h_line_done	; 23   no, we're in vsync, so don't DC restore

	tris	porta		; 24 start DC restore, 2.0 us pulse
	if	dcrest_dbg
	bsf	portb,dcd	; 25
	else
	nop			; 25
	endif

	call	delay7		; 26-32

	movlw	tris_a_no_dc_rest ; 33 prepare to end DC restore

	if	dcrest_dbg
	bcf	portb,dcd	; 34
	else
	nop			; 34
	endif
	tris	porta		; 35 end DC restore

	goto	h_line_done	; 36

; --- rest of interrupt routines follow the tables ---


;-----------------------------------------------------------------------------
; Copyright message
;-----------------------------------------------------------------------------

vermsg:	addwf	pcl
	dt	cr,lf
	dt	"Copyright 1994-1997, 2003-2004 Eric Smith & Richard Ottosen",cr,lf
	dt	"$Id: cc.asm,v 1.23 2004/01/04 22:12:46 eric Exp eric $",cr,lf
	dt	0


;-----------------------------------------------------------------------------
; voting table
;-----------------------------------------------------------------------------

	if	majority

; table to take the majority of the middle three samples
votetab:
	addwf	pcl
	retlw	0	; 00000
	retlw	0	; 00001
	retlw	0	; 00010
	retlw	0	; 00011
	retlw	0	; 00100
	retlw	0	; 00101
	retlw	1	; 00110
	retlw	1	; 00111
	retlw	0	; 01000
	retlw	0	; 01001
	retlw	1	; 01010
	retlw	1	; 01011
	retlw	1	; 01100
	retlw	1	; 01101
	retlw	1	; 01110
	retlw	1	; 01111
	retlw	0	; 10000
	retlw	0	; 10001
	retlw	0	; 10010
	retlw	0	; 10011
	retlw	0	; 10100
	retlw	0	; 10101
	retlw	1	; 10110
	retlw	1	; 10111
	retlw	0	; 11000
	retlw	0	; 11001
	retlw	1	; 11010
	retlw	1	; 11011
	retlw	1	; 11100
	retlw	1	; 11101
	retlw	1	; 11110
	retlw	1	; 11111

	else

; table to look at middle (third) sample only
votetab:
	addwf	pcl
	retlw	0	; 00000
	retlw	0	; 00001
	retlw	0	; 00010
	retlw	0	; 00011
	retlw	1	; 00100
	retlw	1	; 00101
	retlw	1	; 00110
	retlw	1	; 00111
	retlw	0	; 01000
	retlw	0	; 01001
	retlw	0	; 01010
	retlw	0	; 01011
	retlw	1	; 01100
	retlw	1	; 01101
	retlw	1	; 01110
	retlw	1	; 01111
	retlw	0	; 10000
	retlw	0	; 10001
	retlw	0	; 10010
	retlw	0	; 10011
	retlw	1	; 10100
	retlw	1	; 10101
	retlw	1	; 10110
	retlw	1	; 10111
	retlw	0	; 11000
	retlw	0	; 11001
	retlw	0	; 11010
	retlw	0	; 11011
	retlw	1	; 11100
	retlw	1	; 11101
	retlw	1	; 11110
	retlw	1	; 11111

	endif


;-----------------------------------------------------------------------------
; command table
;-----------------------------------------------------------------------------

; This table must be in the first 256 words of the EPROM or the command
; scanner will have to be modified to set the PCLATH register.  It shouldn't
; cross a page boundary or the scanner will have to be modified a lot.

cmd_tbl:
	retlw	'D'
	goto	dbgcmd
	retlw	'N'
	goto	normcmd
	retlw	'G'
	goto	gocmd
	retlw	'R'
	goto	rawcmd
	retlw	'E'
	goto	errcmd
	retlw	'F'		; enable frame count
	goto	frmcmd
	retlw	'C'		; set frame count
	goto	cntcmd
	retlw	'V'
	goto	vercmd
	retlw	0
	goto	badcmd


;-----------------------------------------------------------------------------
; control code table
;-----------------------------------------------------------------------------

cntltbl:
	retlw	2dh
	goto	docr
	retlw	0
	goto	dounkn


;-----------------------------------------------------------------------------
; more of interrupt handler
;-----------------------------------------------------------------------------

vsync_start:
	movwf	prev_portb	; 15

	btfss	portb,oddf
	goto	vsync_start_even

vsync_start_odd:
	bsf	intflags,odd_field
	bcf	intflags,even_field

	movlw	5
	movwf	line_cnt

	btfsc	intflags,hold_time_code	; is main changing the frame counter?
	goto	check_uart	; yes, don't increment it

	call	incr_time_code
	goto	check_uart


vsync_start_even:
	bcf	intflags,odd_field
	bsf	intflags,even_field
	movlw	4
	movwf	line_cnt


h_line_done:
	movf	line_cnt,w	; line 21 or later?
	addlw	256-21
	btfsc	status,c
	goto	check_uart	; yes, don't worry about half-lines

; delay more than half a scan line (31.778 us), to skip any
; half-line serration or equalizing pulse

	movlw	256-40
half_line_delay:
	addlw	1
	btfss	status,c
	goto	half_line_delay

	bcf	intcon,intf	;  clear interrupt
	fallto	check_uart


check_uart:
	btfsc	pir1,rcif	; UART receive?
	call	rx_int		; max 19 cycles

	btfsc	pir1,txif	; UART transmit?
	call	tx_int		; max 18 cycles

	movlw	pie1
	movwf	fsr
	bcf	indf,tmr1ie

	movlw	0fch		; reset csync timeout (772 instr, 154.4 us)
	movwf	tmr1l
	movwf	tmr1h

	bcf	pir1,tmr1if
	bsf	indf,tmr1ie

	goto	irq_done


;-----------------------------------------------------------------------------
; line 21: collect caption data samples for the full scan line
;-----------------------------------------------------------------------------

line21:
	movf	intflags,w	; 23 are we supposed to capture this field
	andwf	capflags,w	; 24    (even or odd)?
	btfsc	status,z	; 25
	goto	no_cap		; 26 no, so skip it

	if	samp_dbg
	nop			; 27
	nop			; 28
	else
	btfsc	sflags,s_ready	; 27 if a sample is already in the buffer,
	goto	overrun		; 28   indicate overrun
	endif

	if	samp_dbg
	bsf	portb,dcd	; 29 pulse to trigger scope
	nop			; 30
	bcf	portb,dcd	; 31
	else
	nop			; 29
	nop			; 30
	nop			; 31
	endif

	movf	tc_frame,w	; 32 record when sample was taken
	movwf	samp_tc_frame	; 33
	movf	tc_sec,w	; 34
	movwf	samp_tc_sec	; 35
	movf	tc_min,w	; 36
	movwf	samp_tc_min	; 37
	movf	tc_hr,w		; 38
	movwf	samp_tc_hr	; 39

	bcf	sflags,s_even_field ; 40 clear sample field flags
	bcf	sflags,s_odd_field  ; 41

	goto	$+1		; 42-43

	movlw	tris_a_no_clamp	; 44 Release peak detect clamp
	tris	porta		; 45

	movf	intflags,w	; 46 remember which field sample was taken from
	movlw	even_field+odd_field	; 47
	iorwf	sflags		; 48

	if	samp_dbg
	bsf	portb,dcd	; 49
	else
	nop			; 49
	endif

scount	set	0
	while	scount<samp_count
	rlf	datap,w		  ; 50 (samp 0) to 304 (samp 127)
	rlf	sample+(scount/8) ; 51 (samp 0) to 305 (samp 127)
scount	set	scount+1
	endw

	if	samp_dbg
	bcf	portb,dcd	; 306
	else
	nop			; 306
	endif

	bsf	sflags,s_ready	; 307

irq_done:			; restore context and return from interrupt
	movf	irq_fsr,w	; 308  (line 21 case)
	movwf	fsr		; 309
	swapf	irq_stat,w	; 310
	movwf	status		; 311
	swapf	irq_w		; 312
	swapf	irq_w,w		; 313
	retfie			; 314-315


overrun:
	bsf	sflags, s_overrun
no_cap:
	goto	irq_done


;-----------------------------------------------------------------------------
; more short delay subroutines
;-----------------------------------------------------------------------------

delay7:	goto	delay5


;-----------------------------------------------------------------------------
; command scanner
;-----------------------------------------------------------------------------

; Enter with W-reg holding the command character and
; "CmdPtr" holding the address of the first byte of table to be scanned.
; The table must be terminated with a null "command" =0 which has its own
; address entry to handle the "command not found" condition.

; The table must be entirely contained in the first 256 words of the EPROM.

cmd_scan:
	movwf	Temp		;Save the command
Cmd10:	movf	CmdPtr,w
	call	ScanGo
	andlw	0FFh		;Command table terminator?
	btfsc	status,z
	goto	Cmd20		; Yes, then do "no command found" and return
	xorwf	Temp,w		;Does table entry match command?
	btfsc	status,z
	goto	Cmd20		; Yes, then do the command found
	incf	CmdPtr		;Point to next command in table
	incf	CmdPtr
	goto	Cmd10

Cmd20:	incf	CmdPtr,w	;Execute the command and return from there
ScanGo:	movwf	pcl		;Get command from table


;-----------------------------------------------------------------------------
; utilities
;-----------------------------------------------------------------------------

	if	has_pzt

; delay for w *1ms at 4mhz osc. freq.
delay:	movwf	mscnt		; count milliseconds
dly10:	movlw	250
	movwf	dlycnt
dly20:	decf	dlycnt		;				1 cycle
	btfss	status,z	;				1
	goto	dly20		;				2 =4 cycles
	decfsz	mscnt		; last msec?
	goto	dly10
	return

; make a "bell" sound
beep:	movlw	200
	movwf	cycles
	movlw	254		; a nice tone???
	movwf	period		; (fall into "click")

; make a "click" sound "cycles" number of times.
;	frequency = 1 /(1 /(osc frequency /4) *5 cycles *period *2)
click:	bsf	portb,pzt	; click high
	movf	period,w	; time for one half of cycle
	movwf	dlycnt		;   into delay counter
clk10:	goto	clk15		; (burn 2 cycles)		2 cycles
clk15:	decfsz	dlycnt		; 				1
	goto	clk10		; 				2= 5 cycles
	bcf	portb,pzt	; click low
	movf	period,w	; time for one half of cycle
	movwf	dlycnt		;   into delay counter
clk20:	goto	clk25		; (burn 2 cycles)		2 cycles
clk25:	decfsz	dlycnt		;				1
	goto	clk20		;				2= 5 cycles

	decfsz	cycles		; another cycle of click?
	goto	click		;   branch if so
	return

	endif


;-----------------------------------------------------------------------------
; time code functions
;-----------------------------------------------------------------------------

incr_time_code:
	incf	tc_frame
	movf	tc_frame,w
	xorlw	30
	btfss	status,z
	return

	clrf	tc_frame
	incf	tc_sec
	movf	tc_sec,w
	xorlw	60
	btfss	status,z
	return

	clrf	tc_sec
	incf	tc_min
	movf	tc_min,w
	xorlw	60
	btfsc	status,z
	goto	incr_hr

; At this point we've advanced the minute to a value between 01 and 59.

	btfsc	mflags,nondrop	; drop-frame time code?
	return			; no, we're done

; We're doing drop-frame time code.  Is the minute a multiple of 10?

	movf	tc_min,w

df_loop:
	addlw	-10		; subtract 10
	btfsc	status,c	; is the result negative?
	return			; yes, must not be a multiple of 10.  done!
	btfss	status,z	; is the result zero?
	goto	df_loop		; no, keep trying

	movlw	2		; yes, set the frame count to 02
	movwf	tc_frame
	return


incr_hr:
	clrf	tc_min
	incf	tc_hr
	movf	tc_hr,w
	xorlw	100
	btfss	status,z
	return

	clrf	tc_hr
	return


reset_time_code:
	bsf	intflags,hold_time_code
	clrf	tc_frame
	clrf	tc_sec
	clrf	tc_min
	clrf	tc_hr
	bcf	intflags,hold_time_code
	return


in_time_code:
; $$$ more code needed here
	return


out_time_code:
	movf	tc_frame,w
	movwf	samp_tc_frame
	movf	tc_sec,w
	movwf	samp_tc_sec
	movf	tc_min,w
	movwf	samp_tc_min
	movf	tc_hr,w
	movwf	samp_tc_hr

	movf	tc_frame,w	; did it change?
	xorwf	samp_tc_frame,w
	btfss	status,z
	goto	out_time_code

	movf	samp_tc_frame,w
	call	outbcd
	movlw	':'
	call	xmit

	movf	samp_tc_sec,w
	call	outbcd
	movlw	':'
	call	xmit

	movf	samp_tc_min,w
	call	outbcd
	movlw	':'
	call	xmit

	movf	samp_tc_hr,w
	fallto	outbcd

outbcd:	movwf	temp
	movlw	-1
	movwf	temp2
	movlw	-10

ob1:	incf	temp2
	addwf	temp,w
	btfsc	status,c
	goto	ob1

	movf	temp2,w
	addlw	'0'
	call	xmit

	movf	temp,w
	addlw	'0'+10
	goto	xmit


;-----------------------------------------------------------------------------
; command to set frame count
;-----------------------------------------------------------------------------

cntcmd:
	call	reset_time_code

; $$$ more code needed here

	return
	

;-----------------------------------------------------------------------------
; output a byte in binary
; uses temp, temp2, and temp3
;-----------------------------------------------------------------------------

	if	0

outbb:	movwf	temp
	movlw	8
	movwf	temp3

outbb1:	movlw	'0'
	btfsc	temp,7
	addlw	1

	call	xmit

	rlf	temp
	decfsz	temp3
	goto	outbb1
	return

	endif


;-----------------------------------------------------------------------------
; output a byte in hexadecimal with braces
; uses temp and temp2
;-----------------------------------------------------------------------------

outhbb:	movwf	temp
outhbbt:
	movlw	'{'
	call	xmit
	call	outhb1
	movlw	'}'
	goto	xmit

;-----------------------------------------------------------------------------
; output a byte in hexadecimal
; uses temp and temp2
;-----------------------------------------------------------------------------

outhb:	movwf	temp
outhb1:	swapf	temp,w
	call	outhd
	movf	temp,w

; fall into outhd

;-----------------------------------------------------------------------------
; output a hex digit
; uses temp2
;-----------------------------------------------------------------------------

outhd:	andlw	0fh
	addlw	0f6h
	btfsc	status,c
	addlw	07h
	addlw	3ah

; fall into xmit

;-----------------------------------------------------------------------------
; output a character, handling lazy CR if necessary
; uses temp2
;-----------------------------------------------------------------------------

xmit:	movwf	temp2		; save the character
	xorlw	cr		; is it a CR?
	btfss	status,z
	goto	xmit0		; no

	bsf	mflags,lazy_cr	; yes, do it later
	return

xmit0:	btfss	mflags,lazy_cr	; is a CR pending?
	goto	xmit1		; no
	
	bcf	mflags,lazy_cr

	movlw	cr		; do the CR
	call	xmit2
	movlw	lf
	call	xmit2

	btfss	mode,fcmode	; frame count?
	goto	xmit1

	call	out_time_code	; output the time code
	movlw	cr		; followed by another CRLF
	call	xmit2
	movlw	lf
	call	xmit2

xmit1:	movf	temp2,w		; restore the character

; fall into xmit2

;-----------------------------------------------------------------------------
; serial transmit character
; blocks until space if available in buffer
;-----------------------------------------------------------------------------

xmit2:	movwf	tx_temp
	movf	fsr,w
	movwf	tx_fsr_save

xmit3:	movf	tx_ring_cnt,w	; is there room for a character?
	xorlw	tx_ring_size
	btfsc	status,z
	goto	xmit3		; no, wait 'til there is

	movf	tx_ring_ip,w	; store char in buffer
	movwf	fsr
	movf	tx_temp,w
	movwf	indf

	ringadv	tx_ring_ip,tx_ring,tx_ring_size	; advance pointer

	incf	tx_ring_cnt	; increment tx char count
				; (this must not be done until after character
				; has been stored in buffer)
	movf	tx_fsr_save,w
	movwf	fsr

	movf	tx_temp,w
	return


;-----------------------------------------------------------------------------
; serial receive character
; non-blocking, returns with carry set if a character was received
;-----------------------------------------------------------------------------

recv:	bcf	status,c
	movf	rx_ring_cnt	; is there anything in buffer?
	btfsc	status,z
	return			; no, return with carry clear

	movf	rx_ring_op,w
	movwf	fsr

	ringadv	rx_ring_op,rx_ring,rx_ring_size	; advance pointer

	movf	indf,w		; get character from buffer

	decf	rx_ring_cnt	; decrement rx char count
				; (this must not be done until after character
				; has been pulled from buffer)

	bsf	status,c
	return


;-----------------------------------------------------------------------------
; UART receive handler
; max 19 cycles including call and return
;-----------------------------------------------------------------------------

rx_int:	movf	rx_ring_cnt,w	; is buffer full?
	xorlw	rx_ring_size
	btfsc	status,z
	goto	rx_overrun

	movf	rx_ring_ip,w	; store character in buffer
	movwf	fsr
	movf	rcreg,w
	movwf	indf

	incf	rx_ring_cnt	; increment count

	ringadv	rx_ring_ip,rx_ring,rx_ring_size		; advance pointer

	return


rx_overrun:
	movf	rcreg,w		; read and discard character
	return


;-----------------------------------------------------------------------------
; UART transmit handler
; max 18 cycles including call and return
;-----------------------------------------------------------------------------

tx_int:	movf	tx_ring_cnt	; is buffer empty?
	btfsc	status,z
	goto	tx_underrun

	movf	tx_ring_op,w	; transmit char
	movwf	fsr
	movf	indf,w
	movwf	txreg

	decf	tx_ring_cnt	; decrement count

	ringadv	tx_ring_op,tx_ring,tx_ring_size		; advance pointer

tx_underrun:			; not interrupt driven, so we don't have
	return			;   to do anything.


;-----------------------------------------------------------------------------
; initialize serial port
;-----------------------------------------------------------------------------

ser_init:
	clrf	rx_ring_cnt
	movlw	rx_ring
	movwf	rx_ring_ip
	movwf	rx_ring_op

	clrf	tx_ring_cnt
	movlw	tx_ring
	movwf	tx_ring_ip
	movwf	tx_ring_op

	movlw	spbrg
	movwf	fsr
	movlw	brg_val
	movwf	indf

	bcf	pir1,rcif	; clear recieve flag

	movlw	txsta
	movwf	fsr
	movlw	(1<<txen)|(1<<brgh)
	movwf	indf

	movlw	(1<<spen)|(1<<cren)
	movwf	rcsta

	return

	
;-----------------------------------------------------------------------------
; PWM init
; generate 50 KHz output for inverter, using timer 2 in PWM mode
;-----------------------------------------------------------------------------

	if	has_pwm_inv

pwm_freq	equ	50000
pwm_prescaler	equ	1
pwm_divisor	equ	fosc/(4*pwm_prescaler*pwm_freq) - 1
pwm_duty	equ	(pwm_divisor+1)*2

pwm_init:
	movlw	pr2
	movwf	fsr
	movlw	pwm_divisor	; set PWM period
	movwf	indf

	movlw	pwm_duty>>2	; set high eight bits of duty cycle
	movwf	ccpr1l

	movlw	((pwm_duty&3)<<ccp1y)|(0ch<<ccp1m0)
	movwf	ccp1con	; set CCP to PWM mode, low two bits of duty cycle

	movlw	(1<<tmr2on)|(0<<t2ckps0)	; start timer 2, 1:1 prescaler
	movwf	t2con

	return

	endif


;-----------------------------------------------------------------------------
; put a character into the command buffer
;-----------------------------------------------------------------------------

put_cmd_char:
	movwf	tx_temp

	movf	cmd_buf_ip,w	; is there room for a character?
	xorlw	cmd_buf+cmd_buf_size
	btfsc	status,z
	return			; no drop it on the floor

	movf	cmd_buf_ip,w	; store char
	movwf	fsr
	movf	tx_temp,w
	movwf	indf

	incf	cmd_buf_ip
	return


;-----------------------------------------------------------------------------
; get a character from the command buffer
; returns with carry clear if no char available
;-----------------------------------------------------------------------------

get_cmd_char:
	bcf	status,c
	movf	cmd_buf_op,w	; is there anything in the buffer?
	xorwf	cmd_buf_ip,w
	btfsc	status,z
	return			; no, return with carry clear

	movf	cmd_buf_op,w	; get char
	movwf	fsr
	movf	indf,w

	incf	cmd_buf_op
	bsf	status,c
	return


;-----------------------------------------------------------------------------
; clear mode
;-----------------------------------------------------------------------------

clrmode:
	bcf	mode,dbgmode
	bcf	mode,rawmode
	bcf	mode,fcmode
	bcf	mode,allerr
	return
	

;-----------------------------------------------------------------------------
; command handlers
;-----------------------------------------------------------------------------

normcmd:
	call	clrmode
	return


dbgcmd:
	call	clrmode
	bsf	mode,dbgmode
	return


rawcmd:
	call	clrmode
	bsf	mode,rawmode
	return


frmcmd:	
	call	clrmode
	call	reset_time_code
	bsf	mode,fcmode
	return


errcmd:
	call	clrmode
	bsf	mode,allerr
	return


gocmd:	bsf	mode,run
badcmd:	return


vercmd:	clrf	temp
ver0:	movf	temp,w
	call	vermsg
	andlw	0ffh
	btfsc	status,z
	return
	call	xmit2
	incf	temp
	goto	ver0


;-----------------------------------------------------------------------------
; reset pointer into sample buffer
;-----------------------------------------------------------------------------

resetptr:
	movlw	sample-1
	movwf	sidx8
	movlw	01h
	movwf	sidx1

	return


;-----------------------------------------------------------------------------
; unget a single sample bit from the sample buffer
; doesn't actually change buffer, just backs the pointer up
;-----------------------------------------------------------------------------

unget:	bcf	status,c	; back up bit position
	rlf	sidx1
	btfss	status,c
	return

	incf	sidx1		; it should be 0, so this will make it 1
	decf	sidx8
	return


;-----------------------------------------------------------------------------
; get a single sample bit from the sample buffer
; return it in the C flag
; return with Z flag set if out of bits
;-----------------------------------------------------------------------------

getsbit:
	bcf	status,c	; advance bit position
	rrf	sidx1
	btfss	status,c
	goto	getsb2
	movlw	80h
	movwf	sidx1
	incf	sidx8

	movf	sidx8,w		; test for end of buffer
	xorlw	sample+samp_bytes
	btfss	status,z
	goto	getsb2

	bsf	status,z	; out of bits
	return

getsb2:
	movf	sidx8,w
	movwf	fsr
	movf	indf,w
	andwf	sidx1,w

	btfss	status,z
	goto	getsb1

	btfsc	mflags,last_bit	; got a zero bit, increment trcnt ifb last_bit
	incf	trcnt		;   was a one

	bcf	mflags,last_bit
	bcf	status,c
	bcf	status,z
	return

getsb1:
	btfss	mflags,last_bit	; got a one bit, increment trcnt if last_bit
	incf	trcnt		;   was a zero
	bsf	mflags,last_bit
	bsf	status,c
	bcf	status,z
	return


;-----------------------------------------------------------------------------
; check parity
;-----------------------------------------------------------------------------

parity:
	movwf	temp
	movlw	8
	movwf	bitcnt
	clrw
par1:	xorwf	temp,w
	rrf	temp
	decfsz	bitcnt
	goto	par1
	andlw	01h
	return


;-----------------------------------------------------------------------------
; output one character as ascii, or as in hexadecimal as [xx] if unprintable
;-----------------------------------------------------------------------------

aout:
	andlw	07fh		; strip parity (sigh)
	movwf	temp

	movf	temp		; don't output nulls
	btfsc	status,z
	return
	
	addlw	0e0h		; values from 00 to 1f show hex
	btfss	status,c
	goto	outhbbt

	movf	temp,w		; 7f show hex
	xorlw	07fh
	btfsc	status,z
	goto	outhbbt

	movf	temp,w
	goto	xmit


;-----------------------------------------------------------------------------
; demodulate the raw samples
;-----------------------------------------------------------------------------

k6	equ	7
k7	equ	8

demod:
	clrf	odata
	clrf	odata+1

	call	resetptr

	clrf	trcnt		; transition counter
	bcf	mflags,last_bit

	movlw	'R'		; set up for no run-in found error
	movwf	errflg

	movlw	k6		; Limit how far we look
	movwf	bitcnt
wu10:	call	getsbit
	btfsc	status,c
	goto	findgap		; Found the first one bit of the run-in
	decfsz	bitcnt
	goto	wu10

	return			; no run-in found

; look for the gap between the leadin and the start bit

findgap:
	movlw	'0'		; set up for "no zero bit" error
	movwf	errflg

pone:	movlw	k7		; count down zero bits
	movwf	bitcnt
pzero:	call	getsbit
	btfsc	status,z
	return			; ran out of bits
	btfsc	status,c
	goto	pone		; darn, it's a one, start counting all over
	decfsz	bitcnt
	goto	pzero

; now find a one bit

	movlw	'1'		; set up for "no one bit" error
	movwf	errflg

fone:	call	getsbit
	btfsc	status,z
	return			; ran out of bits
	btfss	status,c
	goto	fone		; darn, it's a zero, look some more

; test the transition counter to make sure leadin was present

tmin	equ	13
tmax	equ	19

	movlw	'R'		; set up for no run-in found error
	movwf	errflg

	movf	trcnt,w		; is (trcnt >= tmin) && (trcnt <= tmax)?
	addlw	255-tmax
	addlw	(tmax-tmin)+1
	btfss	status,c
	return			; no, no run-in found

	call	unget		; push the first sample of the start bit
				;   back into the sample buffer

	movlw	'B'		; set up for "too few bits" error
	movwf	errflg

	movlw	17		; 2 data bytes of 8 bits each, plus 1 start bit
	movwf	bitcnt		;   The start bit will just get rotated
				;   through and fall off the end of odata.

pl:	clrf	temp		; accumulate the extracted sample bits into
				; temp

	movlw	5		; each data bit has 5 samples
	movwf	temp2

pli:	call	getsbit		; get samples and rotate into temp
	btfsc	status,z
	return
	rlf	temp

	decfsz	temp2
	goto	pli

	movf	temp,w
	call	votetab
	addlw	0ffh		; move decoded bit into carry

	rrf	odata+1
	rrf	odata

	decfsz	bitcnt
	goto	pl

	clrf	errflg		; no error found

	return


;-----------------------------------------------------------------------------
; check the parity
;-----------------------------------------------------------------------------

chkpar:	movlw	'P'		; set up for parity error
	movwf	errflg

	movf	odata,w		; check parity of first byte
	call	parity
	btfsc	status,z
	return

	movf	odata+1,w
	call	parity
	btfsc	status,z
	return

	clrf	errflg		; no error found

	return


;-----------------------------------------------------------------------------
; set DCD based on results of demodulation
;-----------------------------------------------------------------------------

setdcd:
	if	(! samp_dbg) && (! dcrest_dbg)
	movf	errflg,w
	btfss	status,z
	bcf	portb,dcd
	btfsc	status,z
	bsf	portb,dcd
	endif
	return


;-----------------------------------------------------------------------------
; output
;-----------------------------------------------------------------------------


output:
	btfsc	mode,rawmode	; if raw mode, don't check parity
	goto	ignpar

	movf	errflg		; if no error so far, go check the parity
	btfsc	status,z
	call	chkpar

ignpar:	call	setdcd		; update DCD line

	btfss	mode,run	; don't output anything unless run is set
	return

	if	has_dbg
	btfsc	mode,dbgmode	; yes, in debug mode?
	goto	dbgout		; yes, print debugging output
	endif

	movf	errflg,w	; were errors detected?
	btfsc	status,z
	goto	noerr		; no

	btfss	mode,allerr	; display decode errors?
	return			; no

	movlw	'{'
	call	xmit

	movf	errflg,w
	call	xmit

	movf	trcnt,w
	call	outhb

	movlw	'}'
	call	xmit

	return


noerr:	btfsc	mode,rawmode	; raw mode?
	goto	rawout		; yes
	goto	decode		; no error


;-----------------------------------------------------------------------------
; debug mode output
;-----------------------------------------------------------------------------

	if	has_dbg

dbgout:
	call	docr

	movlw	sample		; display the raw sampled bits in hex
	movwf	fsr
	movlw	samp_bytes
	movwf	temp3

dl:	movf	indf,w
	call	outhb
	incf	fsr
	decfsz	temp3
	goto	dl

	movlw	' '
	call	xmit

	movf	trcnt,w		; output transition count
	call	outhb

	movlw	' '
	call	xmit

	call	dohex		; output decoded data bytes

	movf	errflg,w
	btfsc	status,z
	goto	d9

	movlw	' '
	call	xmit

	movf	errflg,w	; output error character
	call	xmit

d9:	goto	docr

	endif


;-----------------------------------------------------------------------------
; raw mode output
;-----------------------------------------------------------------------------

rawout:
	movf	odata,w
	call	xmit
	movf	odata+1,w
	goto	xmit


;-----------------------------------------------------------------------------
; normal output
;-----------------------------------------------------------------------------

decode:
	movlw	7fh
	andwf	odata		; strip parity from first byte
	andwf	odata+1		; strip parity from second byte

	movf	odata,w		; if the first byte is >= 20h, handle
	addlw	0e0h		;   it normally
	btfsc	status,c
	goto	dochar

	movf	odata,w		; if the first byte isn't 14h, throw it
	xorlw	014h		;   away
	btfss	status,z
	return

	movlw	cntltbl
	movwf	cmdptr
	movf	odata+1,w
	goto	cmd_scan


;-----------------------------------------------------------------------------
; action functions for decoded data
;-----------------------------------------------------------------------------

docr:	movlw	cr		; it's a line break
	goto	xmit


dochar:	movf	odata,w
	call	aout
	movf	odata+1,w
	goto	aout

dounkn:	btfsc	mode,showunk	; should unknown control codes be shown in hex?
	goto	dohex		; yes
	bsf	mflags,lazy_cr	; no, turn them into line breaks
	return

dohex:	movf	odata,w
	call	outhbb
	movf	odata+1,w
	goto	outhbb


;-----------------------------------------------------------------------------
; panic - fatal error
;-----------------------------------------------------------------------------

panic:	movlw	'?'
	call	xmit


;-----------------------------------------------------------------------------
; init csync timout (timer 1)
;-----------------------------------------------------------------------------

init_csync_timeout:
	movlw	pie1		; disable interrupt
	movwf	fsr
	bcf	indf,tmr1ie
	
	bcf	t1con,tmr1on	; disable timer

	movlw	0fch		; set timer 1 to interrupt after 772 instr.
	movwf	tmr1l		;   (154.4 us, >2 missed scan lines)
	movwf	tmr1h

	bcf	pir1,tmr1if

	movlw	pie1
	movwf	fsr
	movlw	(1<<tmr1ie)
	movwf	indf

	bsf	indf,tmr1ie

	bsf	intcon,peie	; enable timer 1 interrupts

	movlw	(1<<tmr1on)	; start timer 1, (no prescale)
	movwf	t1con

	return


;-----------------------------------------------------------------------------
; program entry point
;-----------------------------------------------------------------------------

reset:
	clrwdt			; reset watchdog timer

	movlw	(1<<not_rbpu)|(1<<psa)|07h
	option			; not_rbpu=1:    no port B pullups
				; intedg=0:  low to high edge for external interrupt
				; t0cs=0:     internal instruction cycle clock
				; t0se=0
				; psa=1:     prescaler assigned to WDT
				; ps2..0 = 7:  WDT prescaler divide by 128

	clrf	tmr0		; clear TMR0 and prescaler

	movlw	tris_a_no_clamp
	tris	porta
	clrf	porta		; Use port A pins as open drain outputs

	movlw	tris_b_val
	tris	portb
	clrf	portb		; 0 volts on speaker, RS-232 idle

	movlw	06h		; two common reference comparators with
	movwf	cmcon		; outputs

	if	has_pwm_inv
	call	pwm_init	; init PWM for inverter
	endif

	call	ser_init	; init serial port

	clrf	intflags	; clear interrupt flags

	clrf	sflags		; clear sampling flags
	clrf	mflags		; clear misc flags

	clrf	capflags
;	bsf	capflags,cap_odd_field	; capture on odd field only
;	bsf	capflags,cap_even_field	; capture on even field only

	clrf	mode		; clear the mode bits
;	bsf	mode,run	; ... except run

	movlw	22		; don't know what video line we're on, so
	movwf	line_cnt	;   assume we're already past line 21

	call	reset_time_code

	if	has_pzt
	call	beep		; indicate we're alive
	endif


;-----------------------------------------------------------------------------
; enable interupts - we re-enter the main loop here after receiving serial
;                    input
;-----------------------------------------------------------------------------

irqen:	movlw	pie1		; disable peripheral interrupts
	movwf	fsr
	clrf	indf

	bcf	intcon,intf	; clear any pending csync interrupt

	movlw	(1<<inte)	; enable external, timer 1 interrupts
	movwf	intcon

	call	init_csync_timeout	; enable and start csync timeout


	bsf	intcon,gie	; set the global interrupt enable

	if	startup_dbg
	movlw	cr
	call	xmit2
	movlw	lf
	call	xmit2
	movlw	'*'
	call	xmit2
	endif


	bsf	capflags,cap_odd_field	; capture on odd field only
;	bsf	capflags,cap_even_field	; capture on even field only
	bsf	mode,run	; ... except run
	bsf	mode,dbgmode	; and debug mode


	if	dbg_echo
	call	vercmd		; banner message
	endif


;-----------------------------------------------------------------------------
; command done
;-----------------------------------------------------------------------------

cmd_done:
	movlw	cmd_buf		; reset the command buffer pointers
	movwf	cmd_buf_ip
	movwf	cmd_buf_op

	if	dbg_echo	; if debug echo enabled, print a prompt
	movlw	cr
	call	xmit2
	movlw	lf
	call	xmit2
	movlw	'*'
	call	xmit2
	endif


;-----------------------------------------------------------------------------
; main loop
;-----------------------------------------------------------------------------

main:	btfss	sflags,s_ready
	goto	check_rx

	call	demod
	call	output
	
	bcf	sflags,s_ready


check_rx:	; handle serial input
	call	recv		; get a character
	btfss	status,c	; was there one?
	goto	main		;   no, never mind

	xorlw	cr		; is it a carriage return?
	btfsc	status,z
	goto	got_cr		; yes, we've got a complete command
	xorlw	cr		; no, undo the previous xor

	if	dbg_echo
	call	xmit2		; echo character
	endif

	call	put_cmd_char	; store character in command buffer
	goto	main


got_cr:
	movlw	cmd_tbl		; prepare for cmd_scan
	movwf	CmdPtr

	call	get_cmd_char	; get first char of command
	btfss	status,c	; if there isn't one,
	goto	cmd_done	;   we're already done

	call	cmd_scan	; go exit the command

	goto	cmd_done	; done

	end
